//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Types/OwningVector.h
//! @brief     Defines and implements templated class OwningVector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_BASE_TYPES_OWNINGVECTOR_H
#define BORNAGAIN_BASE_TYPES_OWNINGVECTOR_H

#include <cstddef>
#include <utility>
#include <vector>

//! A vector of unique pointers to objects that are cloneable.
//!
//! Equips vector<unique_ptr<T>> with copy constructor.
//! For use with polymorphic objects, or in pimpl idiom.

//! The objects pointed to must posses a clone() function.

template <class T>
class OwningVector {
public:
    OwningVector() = default;
    ~OwningVector() { clear(); }
    OwningVector(const OwningVector& other)
    {
        m_v.reserve(other.size());
        for (T* e : other)
            m_v.emplace_back(e->clone());
    }
    OwningVector& operator=(const OwningVector& other)
    {
        if (this == &other)
            return *this;
        OwningVector ret(other);
        std::swap(m_v, ret.m_v);
        return *this;
    }

    void emplace_back(T* e) { m_v.emplace_back(e); }
    void clear()
    {
        for (T* e : *this)
            delete e;
        m_v.clear();
    }

    size_t size() const { return m_v.size(); }
    bool empty() const { return m_v.empty(); }
    T* const& operator[](int i) const { return m_v.operator[](i); }
    T* const& at(int i) const { return m_v.at(i); }
    const T* back() const { return m_v.back(); }

    using Iterator = typename std::vector<T*>::iterator; // "typename" can be dropped under C++20
    using ConstIterator = typename std::vector<T*>::const_iterator;

    ConstIterator begin() const { return m_v.begin(); }
    ConstIterator end() const { return m_v.end(); }
    Iterator begin() { return m_v.begin(); }
    Iterator end() { return m_v.end(); }

private:
    std::vector<T*> m_v;
};

#endif // BORNAGAIN_BASE_TYPES_OWNINGVECTOR_H
#endif // USER_API
