#!/usr/bin/env python3
"""
Example of simulating a rough sample with a
tanh and Nevot-Croce roughness model using BornAgain.

"""

from matplotlib import pyplot as plt
import bornagain as ba
from bornagain import angstrom, ba_plot as bp, deg


def get_sample(roughness_model):
    """
    Defines sample and returns it
    """

    # create materials
    m_ambient = ba.MaterialBySLD("Ambient", 0, 0)
    m_ti = ba.MaterialBySLD("Ti", -1.9493e-06, 0)
    m_ni = ba.MaterialBySLD("Ni", 9.4245e-06, 0)
    m_substrate = ba.MaterialBySLD("SiSubstrate", 2.0704e-06, 0)

    # create layers
    ambient_layer = ba.Layer(m_ambient)
    ti_layer = ba.Layer(m_ti, 30*angstrom)
    ni_layer = ba.Layer(m_ni, 70*angstrom)
    substrate_layer = ba.Layer(m_substrate)

    # define roughness
    roughness = ba.LayerRoughness(10*angstrom)

    # create multilayer
    multi_layer = ba.MultiLayer()
    multi_layer.addLayer(ambient_layer)
    for _ in range(10):
        multi_layer.addLayerWithTopRoughness(ti_layer, roughness)
        multi_layer.addLayerWithTopRoughness(ni_layer, roughness)
    multi_layer.addLayerWithTopRoughness(substrate_layer, roughness)

    multi_layer.setRoughnessModel(roughness_model)

    return multi_layer


def get_simulation(sample):
    """
    Defines and returns a specular simulation.
    """
    simulation = ba.SpecularSimulation()
    n = bp.simargs['n']
    scan = ba.AlphaScan(1.54*angstrom, n, 0, 2*deg)
    simulation.setScan(scan)
    simulation.setSample(sample)
    return simulation


def simulate(roughness_model, title):
    """
    Runs simulation and returns its result.
    """
    sample = get_sample(roughness_model)
    simulation = get_simulation(sample)
    result = simulation.simulate()
    result.setTitle(title)
    return result


if __name__ == '__main__':
    bp.parse_args(sim_n=500)

    results = [
        simulate(ba.RoughnessModel.NEVOT_CROCE, "Névot-Croce"),
        simulate(ba.RoughnessModel.TANH, "Tanh"),
    ]

    bp.plot_multicurve_specular(results)
