//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/DataItem.cpp
//! @brief     Implements class IntensityDataItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/DataItem.h"
#include "Device/Histo/IOFactory.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/Util/Error.h"

void DataItem::setPowerfield(Powerfield<double>* data)
{
    std::unique_lock<std::mutex> lock(m_update_data_mutex);
    m_data.reset(data);
}

void DataItem::setRawDataVector(std::vector<double> data)
{
    if (m_data->getAllocatedSize() != data.size())
        throw Error("DataItem::setRawDataVector() -> Error. "
                    "Different data size.");
    std::unique_lock<std::mutex> lock(m_update_data_mutex);
    m_data->setRawDataVector(data);
    emitDataChanged();
}

QString DataItem::fileName() const
{
    return getItemValue(P_FILE_NAME).toString();
}

void DataItem::setFileName(const QString& filename)
{
    setItemValue(P_FILE_NAME, filename);
}

QDateTime DataItem::lastModified() const
{
    return m_last_modified;
}

bool DataItem::containsNonXMLData() const
{
    return static_cast<bool>(m_data);
}

bool DataItem::loadData(const QString& projectDir)
{
    const QString filename = fileName(projectDir);
    auto* data = IOFactory::readPowerfield(filename.toStdString());
    if (!data)
        return false;
    setPowerfield(data);
    return true;
}

bool DataItem::saveData(const QString& projectDir)
{
    if (!containsNonXMLData())
        return false;

    std::unique_lock<std::mutex> lock(m_update_data_mutex);
    std::unique_ptr<Powerfield<double>> clone(getPowerfield()->clone());
    lock.unlock();
    IOFactory::writePowerfield(*clone, fileName(projectDir).toStdString());
    return true;
}

void DataItem::setLastModified(const QDateTime& dtime)
{
    m_last_modified = dtime;
}

QString DataItem::selectedCoords() const
{
    return getItemValue(P_AXES_UNITS).value<ComboProperty>().getValue();
}

void DataItem::setAxesUnits(const ComboProperty& units)
{
    setItemValue(P_AXES_UNITS, units.variant());
}

SessionItem* DataItem::getAxesUnitsItem() const
{
    return getItem(P_AXES_UNITS);
}

SelectionDescriptor<QString> DataItem::axesUnits() const
{
    return SelectionDescriptor<QString>(getItem(P_AXES_UNITS));
}

bool DataItem::isAxesUnitsPropertyName(const QString& name)
{
    return name == P_AXES_UNITS;
}

DataItem::DataItem(const QString& modelType)
    : SessionItem(modelType)
{
    // name of the file used to serialize given IntensityDataItem
    addProperty(P_FILE_NAME, "undefined");

    addProperty(P_AXES_UNITS, ComboProperty::fromList({"nbins"}).variant());

    mapper()->setOnPropertyChange([this](const QString& name) {
        if (name == P_FILE_NAME)
            setLastModified(QDateTime::currentDateTime());
    });

    mapper()->setOnValueChange([this]() {
        // Powerfield was modified
        setLastModified(QDateTime::currentDateTime());
    });
}
