//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentView.cpp
//! @brief     Implements class InstrumentView
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentView.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Common/StyledToolbar.h"
#include "GUI/View/Instrument/DepthProbeInstrumentEditor.h"
#include "GUI/View/Instrument/GISASInstrumentEditor.h"
#include "GUI/View/Instrument/InstrumentListView.h"
#include "GUI/View/Instrument/OffSpecularInstrumentEditor.h"
#include "GUI/View/Instrument/SpecularInstrumentEditor.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"
#include <QBoxLayout>
#include <QFormLayout>
#include <QGroupBox>
#include <QLabel>
#include <QLineEdit>
#include <QScrollArea>
#include <QTextEdit>

InstrumentView::InstrumentView(QWidget* parent, ProjectDocument* document)
    : QWidget(parent)
    , m_document(document)
{
    setAttribute(Qt::WA_StyledBackground, true);
    setProperty("stylable", true); // for stylesheet addressing

    auto* horizontalLayout = new QHBoxLayout;
    m_instrumentListView = new InstrumentListView(document, this);
    horizontalLayout->addWidget(m_instrumentListView);

    m_scrollArea = new QScrollArea(this);
    m_scrollArea->setWidgetResizable(true);
    m_scrollArea->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    horizontalLayout->addWidget(m_scrollArea, 1);

    auto* toolbar = new StyledToolbar(this);
    toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);
    toolbar->addActions(m_instrumentListView->toolbarActions());

    auto* mainLayout = new QVBoxLayout(this);
    mainLayout->setMargin(0);
    mainLayout->setSpacing(0);
    mainLayout->addWidget(toolbar);
    mainLayout->addLayout(horizontalLayout);

    createWidgetsForCurrentInstrument();
    updateSingleInstrumentMode();

    connect(m_instrumentListView, &InstrumentListView::instrumentSelected, this,
            &InstrumentView::createWidgetsForCurrentInstrument);

    connect(m_document, &ProjectDocument::singleInstrumentModeChanged, this,
            &InstrumentView::updateSingleInstrumentMode);
}

void InstrumentView::showEvent(QShowEvent*)
{
    // disconnect because when this view is visible, no other instance is modifying instruments. By
    // disconnecting, no additional logic is necessary to avoid recursive calls (recursive since
    // this view also causes instrumentChanged to be emitted).
    disconnect(m_document->instrumentsEditController(),
               &InstrumentsEditController::instrumentChanged, this,
               &InstrumentView::onInstrumentChangedFromExternal);
}

void InstrumentView::hideEvent(QHideEvent*)
{
    // when the instrument view gets hidden (meaning another view is shown), it's necessary to
    // listen to changes (e.g. performed by the LinkInstrumentManager).
    connect(m_document->instrumentsEditController(), &InstrumentsEditController::instrumentChanged,
            this, &InstrumentView::onInstrumentChangedFromExternal);
}

void InstrumentView::createWidgetsForCurrentInstrument()
{
    auto* currentInstrument = m_instrumentListView->currentInstrument();
    if (!currentInstrument) {
        m_scrollArea->setWidget(new QWidget(m_scrollArea)); // blank widget
        return;
    }

    QWidget* w = new QWidget(m_scrollArea);
    auto* layout = new QVBoxLayout(w);

    w->setAttribute(Qt::WA_StyledBackground, true);
    w->setProperty("stylable", true); // for stylesheet addressing

    auto* g = new QGroupBox(m_scrollArea);
    g->setTitle(QString("Information (%1 instrument)").arg(currentInstrument->instrumentType()));

    auto* formLayout = new QFormLayout(g);
    formLayout->setMargin(17);
    formLayout->setSpacing(8);
    layout->addWidget(g);

    auto* nameEdit = new QLineEdit(g);
    formLayout->addRow("Name:", nameEdit);
    nameEdit->setText(currentInstrument->instrumentName());
    connect(nameEdit, &QLineEdit::textEdited, this, &InstrumentView::onInstrumentNameEdited);

    auto* descriptionEdit = new QTextEdit(g);
    descriptionEdit->setMinimumWidth(300);
    descriptionEdit->setMaximumHeight(100);
    descriptionEdit->setAcceptRichText(false);
    descriptionEdit->setTabChangesFocus(true);
    descriptionEdit->setPlainText(currentInstrument->description());
    formLayout->addRow("Description:", descriptionEdit);
    connect(descriptionEdit, &QTextEdit::textChanged,
            [=]() { onInstrumentdescriptionEdited(descriptionEdit->toPlainText()); });

    GroupBoxCollapser::installIntoGroupBox(g);

    if (auto* sp = dynamic_cast<SpecularInstrumentItem*>(currentInstrument)) {
        auto* editor =
            new SpecularInstrumentEditor(m_scrollArea, sp, m_document->instrumentsEditController());
        connect(editor, &SpecularInstrumentEditor::dataChanged, this,
                &InstrumentView::onInstrumentChangedByEditor);
        layout->addWidget(editor);
    } else if (auto* os = dynamic_cast<OffSpecularInstrumentItem*>(currentInstrument)) {
        auto* editor = new OffSpecularInstrumentEditor(m_scrollArea, os);
        connect(editor, &OffSpecularInstrumentEditor::dataChanged, this,
                &InstrumentView::onInstrumentChangedByEditor);
        layout->addWidget(editor);
    } else if (auto* gisas = dynamic_cast<GISASInstrumentItem*>(currentInstrument)) {
        auto* editor = new GISASInstrumentEditor(m_scrollArea, gisas);
        connect(editor, &GISASInstrumentEditor::dataChanged, this,
                &InstrumentView::onInstrumentChangedByEditor);
        layout->addWidget(editor);
    } else if (auto* dp = dynamic_cast<DepthProbeInstrumentItem*>(currentInstrument)) {
        auto* editor = new DepthProbeInstrumentEditor(m_scrollArea, dp);
        connect(editor, &DepthProbeInstrumentEditor::dataChanged, this,
                &InstrumentView::onInstrumentChangedByEditor);
        layout->addWidget(editor);
    } else
        ASSERT(false);

    m_scrollArea->setWidget(w);
}

void InstrumentView::onInstrumentNameEdited(const QString& newName)
{
    auto* currentInstrument = m_instrumentListView->currentInstrument();
    if (currentInstrument && currentInstrument->instrumentName() != newName)
        m_document->instrumentsEditController()->setInstrumentName(currentInstrument, newName);
}

void InstrumentView::onInstrumentdescriptionEdited(const QString& t)
{
    auto* currentInstrument = m_instrumentListView->currentInstrument();
    if (currentInstrument && currentInstrument->description() != t) {
        currentInstrument->setDescription(t);
        onInstrumentChangedByEditor();
    }
}

void InstrumentView::onInstrumentChangedByEditor()
{
    m_document->instrumentsEditController()->notifyInstrumentChanged(
        m_instrumentListView->currentInstrument());
}

void InstrumentView::onInstrumentChangedFromExternal(const InstrumentItem* instrument)
{
    if (instrument == m_instrumentListView->currentInstrument())
        createWidgetsForCurrentInstrument();
}

void InstrumentView::updateSingleInstrumentMode()
{
    m_instrumentListView->setVisible(!m_document->singleInstrumentMode());
}
