//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Job/JobResultsPresenter.cpp
//! @brief     Implements class JobResultsPresenter
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Job/JobResultsPresenter.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/View/Job/JobViewActivities.h"
#include "GUI/View/Plot2D/IntensityDataWidget.h"
#include "GUI/View/PlotComparison/FitComparisonWidget.h"
#include "GUI/View/PlotComparison/FitComparisonWidget1D.h"
#include "GUI/View/PlotSpecular/SpecularDataWidget.h"
#include "GUI/View/Projection/IntensityDataProjectionsWidget.h"

JobResultsPresenter::JobResultsPresenter(QWidget* parent)
    : ItemComboWidget(parent)
{
    setWindowTitle("Job Powerfield");
    setObjectName("JobResultsPresenter");

    registerWidget("Heat Map", create_new<IntensityDataWidget>);
    registerWidget("Projections", create_new<IntensityDataProjectionsWidget>);
    registerWidget("Fit 1D Data", create_new<FitComparisonWidget1D>);
    registerWidget("Fit 2D Data", create_new<FitComparisonWidget>);
    registerWidget("Reflectometry", create_new<SpecularDataWidget>);

    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    setMinimumSize(600, 600);
}

QString JobResultsPresenter::itemPresentation() const
{
    if (!currentItem())
        return {};

    const QString value = dynamic_cast<const JobItem*>(currentItem())->presentationType();
    return value.isEmpty() ? selectedPresentation() : value;
}

void JobResultsPresenter::setItem(SessionItem* item)
{
    ItemComboWidget::setItem(item);
}

void JobResultsPresenter::setPresentation(const QString& presentationType)
{
    ItemComboWidget::setPresentation(presentationType);
    auto* item = dynamic_cast<JobItem*>(currentItem());
    if (item && !presentationType.isEmpty())
        item->setPresentationType(presentationType);
}

void JobResultsPresenter::setPresentation(JobViewActivity activity)
{
    QString presentation;
    auto* jobItem = dynamic_cast<JobItem*>(currentItem());
    if (jobItem && !jobItem->isRunning()) {
        switch (activity) {
        case JobViewActivity::Fitting:
            presentation = fitPresentationOfInstrument(jobItem->instrumentItem());
            break;

        case JobViewActivity::RealTime:
        case JobViewActivity::JobView:
            presentation = defaultPresentationOfInstrument(jobItem->instrumentItem());
            break;
        }
        setPresentation(presentation);
    }
}

//! Returns list of presentation types, available for given item. JobItem with fitting abilities
//! is valid for IntensityDataWidget and FitComparisonWidget.

QStringList JobResultsPresenter::activePresentationList(SessionItem* item)
{
    auto* jobItem = dynamic_cast<JobItem*>(item);
    if (!jobItem || jobItem->isRunning())
        return {};

    ASSERT(jobItem->instrumentItem() != nullptr);

    auto result = defaultActivePresentationsOfInstrument(jobItem->instrumentItem());

    if (jobItem->isValidForFitting())
        result << fitPresentationOfInstrument(jobItem->instrumentItem());

    return result;
}

QStringList JobResultsPresenter::presentationList(SessionItem* item)
{
    auto* jobItem = dynamic_cast<JobItem*>(item);
    if (!jobItem || jobItem->isRunning())
        return {};

    ASSERT(jobItem->instrumentItem() != nullptr);

    auto result = defaultActivePresentationsOfInstrument(jobItem->instrumentItem());
    auto addon = fitPresentationOfInstrument(jobItem->instrumentItem());
    if (!addon.isEmpty())
        result << addon;

    return result;
}

QString JobResultsPresenter::defaultPresentationOfInstrument(InstrumentItem* instrumentItem)
{
    if (instrumentItem->is<SpecularInstrumentItem>())
        return "Reflectometry";

    return "Heat Map";
}

QStringList
JobResultsPresenter::defaultActivePresentationsOfInstrument(InstrumentItem* instrumentItem)
{
    if (instrumentItem->is<SpecularInstrumentItem>())
        return {"Reflectometry"};

    return {"Heat Map", "Projections"};
}

QString JobResultsPresenter::fitPresentationOfInstrument(InstrumentItem* instrumentItem)
{
    if (instrumentItem->is<SpecularInstrumentItem>())
        return "Fit 1D Data";

    return "Fit 2D Data";
}
