//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/SampleListModel.cpp
//! @brief     Implements class SampleListModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/SampleListModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/MultiLayerItems.h"
#include "GUI/Util/String.h"
#include "GUI/View/FromDomain/GUIExamplesFactory.h"
#include "GUI/View/FromDomain/GUISampleBuilder.h"
#include "GUI/View/Global/Globals.h"
#include "GUI/View/Info/DetailedMessageBox.h"
#include "GUI/View/Project/PyImportAssistant.h"
#include "Sample/Multilayer/MultiLayer.h"
#include <QIcon>

SampleListModel::SampleListModel(QObject* parent, MultiLayerItems* model)
    : QAbstractListModel(parent)
    , m_multiLayerItems(model)
{
}

int SampleListModel::rowCount(const QModelIndex& parent) const
{
    if (parent.isValid())
        return 0;

    return m_multiLayerItems->multiLayerItems().size();
}

QVariant SampleListModel::data(const QModelIndex& index, int role) const
{
    auto* const item = itemForIndex(index);

    if (role == Qt::ToolTipRole)
        return item->description();

    if (role == Qt::DisplayRole) {
        auto descr = item->description();
        if (!descr.isEmpty()) {
            descr.prepend("<br><br>");
            const int maxDescriptionLines = 8;
            while (descr.count("\n") >= maxDescriptionLines) {
                descr.truncate(descr.lastIndexOf("\n"));
                descr += " [...]";
            }
            descr.replace("\n", "<br>");
        }
        return "<b>" + item->sampleName() + "</b>" + descr;
    }

    if (role == Qt::EditRole)
        return item->sampleName();

    return QVariant();
}

Qt::ItemFlags SampleListModel::flags(const QModelIndex& index) const
{
    auto f = QAbstractItemModel::flags(index);
    f |= Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsDragEnabled;

    return f;
}

bool SampleListModel::setData(const QModelIndex& index, const QVariant& value, int role)
{
    if (!index.isValid())
        return false;

    if (role == Qt::EditRole && index.column() == 0) {
        itemForIndex(index)->setSampleName(value.toString());
        emit dataChanged(index, index);
        return true;
    }

    if (role == Qt::ToolTipRole && index.column() == 0) {
        itemForIndex(index)->setDescription(value.toString());
        emit dataChanged(index, index);
        return true;
    }

    return false;
}

MultiLayerItem* SampleListModel::itemForIndex(const QModelIndex& index) const
{
    if (!index.isValid())
        return nullptr;

    return m_multiLayerItems->multiLayerItems()[index.row()];
}

QModelIndex SampleListModel::indexForItem(MultiLayerItem* item) const
{
    if (auto row = m_multiLayerItems->multiLayerItems().indexOf(item); row >= 0)
        return index(row, 0);

    return QModelIndex();
}

void SampleListModel::removeSample(MultiLayerItem* item)
{
    QModelIndex index = indexForItem(item);
    if (!index.isValid())
        return;

    beginRemoveRows(index.parent(), index.row(), index.row());
    m_multiLayerItems->removeMultiLayer(item);
    endRemoveRows();
}

QModelIndex SampleListModel::createSample()
{
    const QStringList existingNames = m_multiLayerItems->multiLayerNames();

    const int row = m_multiLayerItems->multiLayerItems().size();
    beginInsertRows(QModelIndex(), row, row);
    auto* multilayer_item = m_multiLayerItems->addMultiLayer();
    multilayer_item->setSampleName(GUI::Util::String::suggestName(existingNames, "Sample"));
    multilayer_item->addStandardMaterials();
    endInsertRows();
    return indexForItem(multilayer_item);
}

QModelIndex SampleListModel::createSampleFromExamples(const QString& className,
                                                      const QString& title,
                                                      const QString& description)
{
    try {
        auto* sample =
            dynamic_cast<MultiLayerItem*>(GUI::ExamplesFactory::createMultiLayerItem(className));
        if (!sample)
            return {};
        sample->setSampleName(title);
        sample->setDescription(description);

        const int row = m_multiLayerItems->multiLayerItems().size();
        beginInsertRows(QModelIndex(), row, row);
        m_multiLayerItems->addMultiLayer(sample);
        endInsertRows();
        return indexForItem(sample);
    } catch (const GUI::Transform::FromDomain::GUISampleBuilder::Exception& ex) {
        QString message("Exception thrown while trying to build GUI models.\n");
        QString details = QString::fromStdString(std::string(ex.what()));
        DetailedMessageBox(GUI::Global::mainWindow, "Example builder failure", message, details)
            .exec();
        return {};
    }
}

#ifdef BORNAGAIN_PYTHON
QModelIndex SampleListModel::createSampleFromPython()
{
    auto multiLayer = PyImportAssistant::importMultiLayer();
    if (!multiLayer)
        return {}; // any messages already shown to user; no dlg necessary anymore


    auto* sample = PyImportAssistant::createMultiLayerItem(*multiLayer);
    if (!sample)
        return {}; // any messages already shown to user; no dlg necessary anymore

    sample->setDescription("Imported from python code");

    const int row = m_multiLayerItems->multiLayerItems().size();
    beginInsertRows(QModelIndex(), row, row);
    m_multiLayerItems->addMultiLayer(sample);
    endInsertRows();
    return indexForItem(sample);
}
#endif
