//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/FullSphere.cpp
//! @brief     Implements class FullSphere.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/FullSphere.h"
#include "Base/Math/Constants.h"
#include "Sample/HardParticle/TruncatedSphere.h"
#include "Sample/LibFF/SomeFormFactors.h"
#include "Sample/Scattering/Rotations.h"

FullSphere::FullSphere(const std::vector<double> P, bool position_at_center)
    : IFormFactor(P)
    , m_radius(m_P[0])
    , m_position_at_center(position_at_center)
{
    checkNodeArgs();
}

FullSphere::FullSphere(double radius, bool position_at_center)
    : FullSphere(std::vector<double>{radius}, position_at_center)
{
}

double FullSphere::bottomZ(const IRotation* rotation) const
{
    if (m_position_at_center)
        return -m_radius;
    if (!rotation)
        return 0;
    R3 centre(0.0, 0.0, m_radius);
    R3 new_centre = rotation->transformed(centre);
    return new_centre.z() - m_radius;
}

double FullSphere::topZ(const IRotation* rotation) const
{
    if (m_position_at_center)
        return m_radius;
    if (!rotation)
        return 2 * m_radius;
    R3 centre(0.0, 0.0, m_radius);
    R3 new_centre = rotation->transformed(centre);
    return new_centre.z() + m_radius;
}

complex_t FullSphere::formfactor_at_bottom(C3 q) const
{
    complex_t result = SampleUtils::someff::ffSphere(q, m_radius);
    if (!m_position_at_center)
        result *= exp_I(q.z() * m_radius);
    return result;
}
