//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/OffSpecularSimulation.cpp
//! @brief     Implements class OffSpecularSimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sim/Simulation/OffSpecularSimulation.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Base/Util/Assert.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "Device/Histo/SimulationResult.h"
#include "Param/Distrib/Distributions.h"
#include "Resample/Element/DiffuseElement.h"

OffSpecularSimulation::OffSpecularSimulation(const Beam& beam, const MultiLayer& sample,
                                             const IDetector& detector)
    : ISimulation2D(beam, sample, detector)
{
}

OffSpecularSimulation::OffSpecularSimulation()
{
    initialize();
}

void OffSpecularSimulation::prepareSimulation()
{
    checkInitialization();
    ISimulation2D::prepareSimulation();
}

size_t OffSpecularSimulation::numberOfElements() const
{
    checkInitialization();
    return ISimulation2D::numberOfElements() * m_alpha_i_axis->size();
}

SimulationResult OffSpecularSimulation::pack_result()
{
    checkInitialization();
    const IAxis& phi_axis = detector().axis(0);
    size_t phi_f_size = phi_axis.size();
    ASSERT(phi_f_size * m_intensity_map.getAllocatedSize() == m_eles.size());
    for (size_t i = 0; i < m_alpha_i_axis->size(); ++i)
        transferDetectorImage(i);

    auto data = std::unique_ptr<Powerfield<double>>(m_intensity_map.clone());
    std::unique_ptr<ICoordSystem> coordsSystem(createCoordSystem());
    return {*data, *coordsSystem};
}

void OffSpecularSimulation::setBeamParameters(double wavelength, const IAxis& alpha_axis,
                                              double phi_i)
{
    m_alpha_i_axis.reset(alpha_axis.clone());
    if (alpha_axis.size() < 1)
        throw std::runtime_error("OffSpecularSimulation::prepareSimulation() "
                                 "-> Error. Incoming alpha range size < 1.");
    const double alpha_zero = alpha_axis.lowerBound();
    instrument().setBeamParameters(wavelength, alpha_zero, phi_i);
    updateIntensityMap();
}

const IAxis* OffSpecularSimulation::beamAxis() const
{
    return m_alpha_i_axis.get();
}

#ifndef SWIG
ICoordSystem* OffSpecularSimulation::createCoordSystem() const
{
    ASSERT(beamAxis());

    const auto axes = detector2D().axesClippedToRegionOfInterest();
    ASSERT(axes.size() == 2);
    const IAxis& yAxis = *axes[1];

    if (const auto* rect_det = dynamic_cast<const RectangularDetector*>(&detector2D())) {
        std::unique_ptr<RectangularPixel> det_pixel(rect_det->regionOfInterestPixel());
        return OffSpecularCoordinates::createForRectangularDetector(beam().direction(), *beamAxis(),
                                                                    *det_pixel, yAxis);
    }
    if (dynamic_cast<const SphericalDetector*>(&detector2D()))
        return OffSpecularCoordinates::createForSphericalDetector(beam().direction(), *beamAxis(),
                                                                  yAxis);

    ASSERT(0);
}
#endif

size_t OffSpecularSimulation::intensityMapSize() const
{
    checkInitialization();
    return m_alpha_i_axis->size() * detector().axis(1).size();
}

void OffSpecularSimulation::initElementVector()
{
    m_eles.clear();
    Beam beam2 = beam();
    for (size_t i = 0; i < m_alpha_i_axis->size(); ++i) {
        beam2.setInclination(m_alpha_i_axis->bin(i).center());
        std::vector<std::unique_ptr<DiffuseElement>> eles_i = generateElements(beam2);
        for (auto&& ele : eles_i)
            m_eles.emplace_back(std::move(ele));
    }
    if (m_cache.empty())
        m_cache.resize(m_eles.size(), 0.0);
}

void OffSpecularSimulation::validateParametrization(const ParameterDistribution& par_distr) const
{
    const bool zero_mean = par_distr.getDistribution()->mean() == 0.0;
    if (zero_mean)
        return;
    /* TODO come back to this after refactoring of distribution handling
    if (par_distr.whichParameter() == ParameterDistribution::BeamInclinationAngle)
        throw std::runtime_error("Error in OffSpecularSimulation: parameter distribution of "
                                 "beam inclination angle should have zero mean.");
    */
}

void OffSpecularSimulation::updateIntensityMap()
{
    m_intensity_map.clear();
    if (m_alpha_i_axis)
        m_intensity_map.addAxis(*m_alpha_i_axis);
    size_t detector_dimension = detector().dimension();
    if (detector_dimension == 2)
        m_intensity_map.addAxis(detector().axis(1));
    m_intensity_map.setAllTo(0.);
}

void OffSpecularSimulation::transferDetectorImage(size_t index)
{
    Powerfield<double> detector_image;
    size_t detector_dimension = detector().dimension();
    for (size_t dim = 0; dim < detector_dimension; ++dim)
        detector_image.addAxis(detector().axis(dim));
    size_t detector_size = detector_image.getAllocatedSize();
    for (size_t i = 0; i < detector_size; ++i)
        detector_image[i] = m_eles[index * detector_size + i]->intensity();
    detector().applyDetectorResolution(&detector_image);
    size_t y_axis_size = detector().axis(1).size();
    for (size_t i = 0; i < detector_size; ++i)
        m_intensity_map[index * y_axis_size + i % y_axis_size] += detector_image[i];
}

void OffSpecularSimulation::checkInitialization() const
{
    ASSERT(m_alpha_i_axis);
    ASSERT(m_alpha_i_axis->size() >= 1);
    ASSERT(detector().dimension() == 2);
}

void OffSpecularSimulation::initialize() {}
