#include "Base/Axis/VariableBinAxis.h"
#include "Device/Data/DataUtils.h"
#include "Tests/GTestWrapper/google_test.h"
#include <algorithm>

class PowerfieldTest : public ::testing::Test {
protected:
    PowerfieldTest();

    Powerfield<int> int_data_0d;
    Powerfield<double> db_data_3d;
    std::vector<size_t> zero_3d_coordinate;
};

PowerfieldTest::PowerfieldTest()
{
    zero_3d_coordinate.push_back(0);
    zero_3d_coordinate.push_back(0);
    zero_3d_coordinate.push_back(0);
    FixedBinAxis angle_axis("angle", 20, 0.0, 0.1);
    FixedBinAxis length_axis("length", 10, 0.0, 0.5);
    FixedBinAxis index_axis("index", 10, 1, 10);

    db_data_3d.addAxis(angle_axis);
    db_data_3d.addAxis(length_axis);
    db_data_3d.addAxis(index_axis);
    Powerfield<double>::iterator it = db_data_3d.begin();
    for (size_t i = 0; i < 2000; ++i) {
        *it = (double)i;
        ++it;
    }
}

TEST_F(PowerfieldTest, SingleElementAfterConstruction)
{
    EXPECT_EQ(1u, int_data_0d.getAllocatedSize());
}

TEST_F(PowerfieldTest, SizeAfterAddingAxes)
{
    EXPECT_EQ(1u, int_data_0d.getAllocatedSize());
    EXPECT_EQ(2000u, db_data_3d.getAllocatedSize());
}

TEST_F(PowerfieldTest, DataInitialization)
{
    std::vector<unsigned> coordinates;
    coordinates.push_back(11);
    coordinates.push_back(4);
    coordinates.push_back(3);
    EXPECT_DOUBLE_EQ((double)1143, db_data_3d[db_data_3d.toGlobalIndex(coordinates)]);
}

TEST_F(PowerfieldTest, isInitialized)
{
    Powerfield<double> data1;
    EXPECT_FALSE(data1.isInitialized());
    data1.addAxis("axis", 10, 0.0, 10.0);
    EXPECT_TRUE(data1.isInitialized());
    data1.clear();
    EXPECT_FALSE(data1.isInitialized());
}

TEST_F(PowerfieldTest, DataCopying)
{
    Powerfield<double> data1;
    Powerfield<double> data2;
    data1.addAxis("axis1", 10, 0., 10.);
    data2.addAxis("axis1", 10, 0., 10.);
    EXPECT_TRUE(data1.hasSameDimensions(data2));
    EXPECT_TRUE(data1.hasSameShape(data2));

    data1.addAxis("axis2", 10, 0., 10.);
    data2.addAxis("axis2", 10, 1., 10.);
    EXPECT_TRUE(data1.hasSameDimensions(data2));
    EXPECT_FALSE(data1.hasSameShape(data2));

    data2.copyFrom(data1);
    EXPECT_TRUE(data1.hasSameDimensions(data2));
    EXPECT_TRUE(data1.hasSameShape(data2));

    data1.addAxis("axis3", 10, 0., 10.);
    data2.addAxis("axis3another", 10, 0., 10.);
    EXPECT_TRUE(data1.hasSameDimensions(data2));
    EXPECT_FALSE(data1.hasSameShape(data2));

    data1.setAllTo(10);
    data2.copyFrom(data1);
    EXPECT_TRUE(data1.totalSum() == data2.totalSum());
    EXPECT_TRUE(data1.hasSameDimensions(data2));
    EXPECT_TRUE(data1.hasSameShape(data2));
    EXPECT_TRUE(data2.hasSameDimensions(data1));
    EXPECT_TRUE(data2.hasSameShape(data1));
    EXPECT_TRUE(data1.hasSameDimensions(data1));
    EXPECT_TRUE(data1.hasSameShape(data1));
    EXPECT_TRUE(data2.hasSameDimensions(data2));
    EXPECT_TRUE(data2.hasSameShape(data2));
}

TEST_F(PowerfieldTest, MaxElement)
{
    Powerfield<double> data;
    data.addAxis("axis1", 10, 0., 10.);
    data.addAxis("axis2", 2, 0., 10.);

    Powerfield<double>::iterator it = data.begin();
    for (size_t i = 0; i < data.getAllocatedSize(); ++i)
        if (i == 10)
            (*it) = 10.0;

    Powerfield<double>::const_iterator cit = std::max_element(data.begin(), data.end());
    EXPECT_EQ(10., (*cit));

    EXPECT_EQ(10., data.max());

    EXPECT_EQ(1., data.normalizedToMaximum().max());
}

// y |
// --------------------------------------------
// 1 | 1   3   5   7   9   11  13  15  17  19 |
// 0 | 0   2   4   6   8   10  12  14  16  18 |
// --------------------------------------------
//   | 0   1   2   3   4   5   6   7   8   9  | x
TEST_F(PowerfieldTest, ValueOfAxis)
{
    Powerfield<double> data;
    data.addAxis("axis1", 10, 0., 10.);
    data.addAxis("axis2", 2, 0., 10.);
    EXPECT_EQ(0.5, data.getAxisValue(0, "axis1"));
    EXPECT_EQ(0.5, data.getAxisValue(1, "axis1"));
    EXPECT_EQ(1.5, data.getAxisValue(2, "axis1"));
    EXPECT_EQ(1.5, data.getAxisValue(3, "axis1"));
    EXPECT_EQ(9.5, data.getAxisValue(18, "axis1"));
    EXPECT_EQ(9.5, data.getAxisValue(19, "axis1"));

    EXPECT_EQ(2.5, data.getAxisValue(0, "axis2"));
    EXPECT_EQ(7.5, data.getAxisValue(1, "axis2"));
    EXPECT_EQ(2.5, data.getAxisValue(2, "axis2"));
    EXPECT_EQ(7.5, data.getAxisValue(3, "axis2"));
    EXPECT_EQ(2.5, data.getAxisValue(18, "axis2"));
    EXPECT_EQ(7.5, data.getAxisValue(19, "axis2"));

    std::vector<double> coordinates = data.getAxesValues(18);
    EXPECT_EQ(9.5, coordinates[0]);
    EXPECT_EQ(2.5, coordinates[1]);
}

// y |
// --------------------------------------------
// 1 | 1   3   5   7   9   11  13  15  17  19 |
// 0 | 0   2   4   6   8   10  12  14  16  18 |
// --------------------------------------------
//   | 0   1   2   3   4   5   6   7   8   9  | x
TEST_F(PowerfieldTest, GetAxisBin)
{
    Powerfield<double> data;
    data.addAxis("axis1", 10, 0., 10.);
    data.addAxis("axis2", 2, 0., 10.);

    EXPECT_EQ(0.5, data.getAxisBin(0, "axis1").center());
    EXPECT_EQ(0.5, data.getAxisBin(1, "axis1").center());
    EXPECT_EQ(9.5, data.getAxisBin(18, "axis1").center());
    EXPECT_EQ(9.5, data.getAxisBin(19, "axis1").center());
    EXPECT_EQ(2.5, data.getAxisBin(0, "axis2").center());
    EXPECT_EQ(7.5, data.getAxisBin(1, "axis2").center());
    EXPECT_EQ(2.5, data.getAxisBin(18, "axis2").center());
    EXPECT_EQ(7.5, data.getAxisBin(19, "axis2").center());
}

TEST_F(PowerfieldTest, SetCleared)
{
    db_data_3d.clear();
    db_data_3d.setAllTo(1.0);
    EXPECT_EQ(db_data_3d[0], 1.0);
}

TEST_F(PowerfieldTest, MixedTypeOperations)
{
    Powerfield<bool> data_bool;
    data_bool.addAxis("axis1", 10, 0.0, 10.0);

    Powerfield<bool> data_double;
    data_double.addAxis("axis1", 10, 0.0, 10.0);

    EXPECT_TRUE(data_bool.hasSameShape(data_double));

    data_double.addAxis("axis2", 10, -1.0, 1.0);
    EXPECT_FALSE(data_bool.hasSameShape(data_double));
}
